import os

from datetime import datetime


def get_files(directory, condition=lambda x: True):
    """
    yield tuples of (filename, path) for files in given `directory`
    that match `condition` (default: all) incl. subdirectories
    """
    return (
        (os.path.splitext(f)[0], os.path.join(d, f))
        for d, _, fnames in os.walk(directory)
        for f in fnames
        if condition(os.path.join(d, f))
    )


def cast(value):
    if not isinstance(value, (str, float, int)):
        return value
    if isinstance(value, str):
        value = value.strip()
    try:
        if float(value) == int(float(value)):
            return int(value)
        return float(value)
    except (TypeError, ValueError):
        try:
            return datetime.fromisoformat(value)
        except ValueError:
            return value



from datetime import datetime


class File:
    def __init__(self, metadir, data, identifier="content_hash"):
        self._metadir = metadir
        self._data = data
        self._identifier = identifier
        self.id = self._data[identifier]

    def __setitem__(self, attr, value):
        self.update(**{attr: value})

    def __getitem__(self, attr, default=None):
        return self._data.get(attr, default)

    def __contains__(self, key):
        return key in self._data

    def update(self, **data):
        """
        bulk attribute update (like dict.update)
        update internal data object and write to state db
        """
        self._data.update(**data)
        self._data["__state_last_updated"] = datetime.now()

    def save(self):
        with self._metadir._state_db as db:
            db["files"].update(self._data, [self._identifier])
        self._metadir.touch("state_last_updated")


class FilesWrapper:
    """
    yield actual `File` objects from dataset table,
    pass other operations through `dataset.table.Table`

    https://dataset.readthedocs.io/en/latest/quickstart.html#reading-data-from-tables
    """

    def __init__(self, table, metadir):
        self._table = table
        self._metadir = metadir

    def __iter__(self):
        for data in self._table:
            yield File(self._metadir, data)

    def __len__(self):
        return len(self._table)

    def __contains__(self, file):
        return bool(self.find_one(id=file["id"]))

    def find(self, *args, **kwargs):
        for data in self._table.find(*args, **kwargs):
            yield File(self._metadir, data)

    def find_one(self, *args, **kwargs):
        data = self._table.find_one(*args, **kwargs)
        if data:
            return File(self._metadir, data)

    def __getattr__(self, attr):
        """
        pass through dataset table funcionality
        """
        return getattr(self._table, attr)



import click
import logging

from mmmeta import settings, mmmeta


log = logging.getLogger(__name__)


@click.group()
@click.option(
    "--metadir",
    default=settings.MMMETA,
    help="Base path for reading meta info and storing state",
    show_default=True,
)
@click.option(
    "--files-root",
    default=settings.MMMETA_FILES_ROOT,
    help="Base path for actual files to generate metadir from",
    show_default=True,
)
@click.pass_context
def cli(ctx, metadir, files_root, invoke_without_command=True):
    logging.basicConfig(level=logging.INFO)
    if not metadir:
        raise click.BadParameter("Missing metadir root")
    if ctx.obj is None:
        ctx.obj = {}
    ctx.obj["m"] = mmmeta(metadir)


@cli.command()
@click.pass_context
def generate(ctx):
    ctx.obj["m"].generate()


@cli.command()
@click.pass_context
def update(ctx):
    ctx.obj["m"].update()


@cli.command()
@click.pass_context
def inspect(ctx):
    meta = ctx.obj["m"]
    for key, value in meta.inspect().items():
        click.echo(f"{key}: {value}")
    click.echo(f"{meta.store.to_string()}")



import json
import logging
from datetime import datetime

from banal import is_listish


log = logging.getLogger(__name__)


def _upsert(db, files, prefix, unique="content_hash"):
    with db as tx:
        # use explicit operations instead of upsert_many to be able to set some
        # more metadata
        existing = [f[unique] for f in db["files"]]
        log.info(f"{len(existing)} exsiting files in `{db}`")
        updated = 0
        added = 0

        for file in files:
            file[f"{prefix}_last_updated"] = datetime.now()
            if file[unique] in existing:
                updated += 1
                tx["files"].update(file, [unique])
            else:
                file[f"{prefix}_added"] = datetime.now()
                tx["files"].insert(file)
                added += 1
    return updated, added


def to_db_json(data):
    # FIXME
    return {k: json.dumps(v) if is_listish(v) else v for k, v in data.items()}


def update_state_db(metadir, unique="content_hash"):
    """
    update remote metadata to local state
    """

    log.info(f"Updating metadata and state for `{metadir}` ...")

    updated, added = _upsert(
        metadir._state_db, metadir._meta_db["files"], "__state", unique
    )

    metadir.touch("state_last_updated")

    log.info(f"Updated {updated} files.")
    log.info(f"Added {added} new files.")
    log.info(f'Now {len(metadir._meta_db["files"])} files in state database.')


def generate_meta_db(filebackend, metadir, replace=False, unique="content_hash"):
    """
    generate or update file metadata
    """

    log.info(f"Updating metadata for `{filebackend}` ...")

    if replace:
        # soft delete current meta.db
        pass

    files = (
        to_db_json(json.load(open(fp)))
        for _, fp in filebackend.get_children(condition=lambda x: x.endswith(".json"))
    )

    updated, added = _upsert(metadir._meta_db, files, "__meta", unique)

    metadir.touch("meta_last_updated")

    log.info(f"Updated {updated} files.")
    log.info(f"Added {added} new files.")
    log.info(f'Now {len(metadir._meta_db["files"])} files in meta database.')
